@@ -231,7 +231,7 @@ void autoLagKickRoutine(std::chrono::steady_clock::time_point now)
 			kickPlayer(i, "Your connection was too laggy.", ERROR_CONNECTION, false);
 			ingame.LagCounter[i] = 0;
 		}
-		else if (LagSecondsCount >= (LagAutoKickSeconds - 3)) {
+		else if (LagSecondsCount == (LagAutoKickSeconds - 3)) {
 			std::string msg = astringf("Auto-kicking player %" PRIu32 " (\"%s\") in %u seconds. (lag)", i, getPlayerName(i, true), (LagAutoKickSeconds - LagSecondsCount));
 			debug(LOG_INFO, "%s", msg.c_str());
 			sendInGameSystemMessage(msg.c_str());
@@ -377,6 +377,8 @@ void autoLobbyNotReadyKickRoutine(std::chrono::steady_clock::time_point now)
 	}
 
 	ingame.lastNotReadyCheck = now;
+	std::array<bool, MAX_CONNECTED_PLAYERS> queuedToKickIdx {};
+
 	for (uint32_t i = 0; i < MAX_CONNECTED_PLAYERS; ++i)
 	{
 		if (!isHumanPlayer(i))
@@ -399,6 +401,19 @@ void autoLobbyNotReadyKickRoutine(std::chrono::steady_clock::time_point now)
 
 		auto totalSecondsNotReady = calculateSecondsNotReadyForPlayer(i, now);
 		if (totalSecondsNotReady >= NotReadyAutoKickSeconds) {
+			// record that player should be kicked, but wait to kick until after processing all slots
+			queuedToKickIdx[i] = true;
+		}
+		else if (!NetPlay.players[i].ready && totalSecondsNotReady >= (NotReadyAutoKickSeconds - 8)) {
+			sendQuickChat(WzQuickChatMessage::INTERNAL_LOCALIZED_LOBBY_NOTICE, realSelectedPlayer, WzQuickChatTargeting::targetAll(), WzQuickChatDataContexts::INTERNAL_LOCALIZED_LOBBY_NOTICE::constructMessageData(WzQuickChatDataContexts::INTERNAL_LOCALIZED_LOBBY_NOTICE::Context::NotReadyKickWarning, i, static_cast<uint32_t>(NotReadyAutoKickSeconds - totalSecondsNotReady)));
+		}
+	}
+
+	// actually process pending kicks
+	for (uint32_t i = 0; i < queuedToKickIdx.size(); ++i)
+	{
+		if (queuedToKickIdx[i])
+		{
 			std::string msg = astringf("Auto-kicking player %" PRIu32 " (\"%s\") because they aren't ready. (Timeout: %u seconds)", i, getPlayerName(i), NotReadyAutoKickSeconds);
 			debug(LOG_INFO, "%s", msg.c_str());
 			sendQuickChat(WzQuickChatMessage::INTERNAL_LOCALIZED_LOBBY_NOTICE, realSelectedPlayer, WzQuickChatTargeting::targetAll(), WzQuickChatDataContexts::INTERNAL_LOCALIZED_LOBBY_NOTICE::constructMessageData(WzQuickChatDataContexts::INTERNAL_LOCALIZED_LOBBY_NOTICE::Context::NotReadyKicked, i, static_cast<uint32_t>(NotReadyAutoKickSeconds)));
@@ -409,9 +424,6 @@ void autoLobbyNotReadyKickRoutine(std::chrono::steady_clock::time_point now)
 			}
 			kickPlayer(i, "You have been removed from the room.\nYou have spent too much time without checking Ready.\n\nIn the future, please check Ready and leave it checked, to avoid delaying games for other players.", ERROR_CONNECTION, false);
 		}
-		else if (!NetPlay.players[i].ready && totalSecondsNotReady >= (NotReadyAutoKickSeconds - 8)) {
-			sendQuickChat(WzQuickChatMessage::INTERNAL_LOCALIZED_LOBBY_NOTICE, realSelectedPlayer, WzQuickChatTargeting::targetAll(), WzQuickChatDataContexts::INTERNAL_LOCALIZED_LOBBY_NOTICE::constructMessageData(WzQuickChatDataContexts::INTERNAL_LOCALIZED_LOBBY_NOTICE::Context::NotReadyKickWarning, i, static_cast<uint32_t>(NotReadyAutoKickSeconds - totalSecondsNotReady)));
-		}
 	}
 }
 
@@ -1430,7 +1442,7 @@ bool shouldProcessMessage(NETQUEUE& queue, uint8_t type)
 				char buf[255];
 				auto senderPlayerIdx = queue.index;
 				debug(LOG_INFO, "Auto kicking player %s, invalid command received: %s", getPlayerName(senderPlayerIdx), messageTypeToString(type));
-				ssprintf(buf, _("Auto kicking player %s, invalid command received: %u"), getPlayerName(senderPlayerIdx, true), type);
+				ssprintf(buf, _("Auto kicking player %s, invalid command received: %u"), getPlayerName(senderPlayerIdx, true), static_cast<unsigned>(type));
 				sendInGameSystemMessage(buf);
 				kickPlayer(queue.index, _("Unauthorized network command"), ERROR_INVALID, false);
 			}
